@@ -54,6 +54,7 @@ struct _LmConnection {
 	GMainContext *context;
 	gchar        *server;
 	gchar        *jid;
+	gchar        *resource;
 	guint         port;
 
 	LmSSL        *ssl;
@@ -210,6 +211,7 @@ connection_free (LmConnection *connection)
 	g_free (connection->server);
 	g_free (connection->jid);
 	g_free (connection->stream_id);
+	g_free (connection->resource);
 
 	if (connection->parser) {
 		lm_parser_free (connection->parser);
@@ -285,7 +287,7 @@ connection_handle_message (LmConnection *connection, LmMessage *m)
 	}
 	else if (lm_message_get_type (m) == LM_MESSAGE_TYPE_STREAM_ERROR) {
 		connection_stream_error (connection, m);
-		goto out;
+		goto run_global_handlers;
 	}
 	
 	id = lm_message_node_get_attribute (m->node, "id");
@@ -304,6 +306,7 @@ connection_handle_message (LmConnection *connection, LmMessage *m)
 		goto out;
 	}
 
+run_global_handlers:
 	for (l = connection->handlers[lm_message_get_type (m)]; 
 	     l && result == LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS; 
 	     l = l->next) {
@@ -1461,7 +1464,7 @@ connection_auth_reply (LmMessageHandler *handler,
 static void
 connection_stream_error (LmConnection *connection, LmMessage *m)
 {
-	LmMessageNode *node;
+	LmMessageNode *node, *child;
 
 	g_return_if_fail (connection != NULL);
 	g_return_if_fail (m != NULL);
@@ -1469,16 +1472,16 @@ connection_stream_error (LmConnection *connection, LmMessage *m)
 	node = m->node;
 
 	/* Resource conflict */
-	node = lm_message_node_get_child (node, "conflict");
-	if (node) {
+	child = lm_message_node_get_child (node, "conflict");
+	if (child) {
 		lm_verbose ("Stream error: Conflict (resource connected elsewhere)\n");
 		connection->disconnect_reason = LM_DISCONNECT_REASON_RESOURCE_CONFLICT;
 		return;
 	}
 
 	/* XML is crack */
-	node = lm_message_node_get_child (node, "xml-not-well-formed");
-	if (node) {
+	child = lm_message_node_get_child (node, "xml-not-well-formed");
+	if (child) {
 		lm_verbose ("Stream error: XML not well formed\n");
 		connection->disconnect_reason = LM_DISCONNECT_REASON_INVALID_XML;
 		return;
@@ -1617,6 +1620,7 @@ lm_connection_new (const gchar *server)
 	connection->context           = NULL;
 	connection->port              = LM_CONNECTION_DEFAULT_PORT;
 	connection->jid               = NULL;
+	connection->resource          = NULL;
 	connection->ssl               = NULL;
 	connection->proxy             = NULL;
 	connection->disconnect_cb     = NULL;
@@ -1863,6 +1867,8 @@ lm_connection_authenticate (LmConnection      *connection,
 	data->password = g_strdup (password);
 	data->resource = g_strdup (resource);
 	
+	/* assume that this is our visible resource. */
+	connection->resource = g_strdup (resource);
 	handler = lm_message_handler_new (connection_auth_req_reply, 
 					  data, 
 					  (GDestroyNotify) auth_req_data_free);
@@ -2074,6 +2080,23 @@ lm_connection_get_jid (LmConnection *connection)
 	return connection->jid;
 }
 
+/**
+ * lm_connection_get_full_jid:
+ * @connection: an #LmConnection
+ *
+ * Creates a full jid for @connection out of the connection jid and the 
+ * resource used for authentication.
+ *
+ * Return value: A newly created string representing the full jid.
+ **/
+gchar *
+lm_connection_get_full_jid (LmConnection *connection)
+{
+	g_return_val_if_fail (connection != NULL, NULL);
+
+	return g_strconcat(connection->jid, "/", connection->resource, NULL);
+}
+
 /**
  * lm_connection_set_jid:
  * @connection: an #LmConnection
@@ -2483,6 +2506,21 @@ lm_connection_send_raw (LmConnection  *connection,
 
 	return connection_send (connection, str, -1, error);
 }
+
+/**
+ * lm_connection_get_client_host:
+ * @connection: An #LmConnection
+ *
+ * Returns the local host name of the connection.
+ *
+ * Return value: A newly allocated string representing the local host name.
+ **/
+gchar *
+lm_connection_get_local_host (LmConnection *connection)
+{
+	return _lm_sock_get_local_host (connection->fd);
+}
+
 /**
  * lm_connection_get_state:
  * @connection: Connection to get state on
